==> heteroscalapack-1.0.2-BETA.tar.gz: heteroscalapack-1.0.2-BETA/tests/RECON/level3/test_hscal_pdgemm_pauto.c <==

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hscalapack.h>

#include "hscalapack_pdgemm_parameters.h"

#define  HSCAL_TEST_NUM_RECON_ITERATIONS  2

static int p, q;

/*------------------------------------------------------------------*/

void hscal_pdgemm_benchmark_function_t
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;

   int TRANSAI       =  params[0];
   int TRANSBI       =  params[1];
   int M             =  params[1+1];
   int N             =  params[1+1+1];
   int K             =  params[1+1+1+1];
   hscaldoublereal ALPHA  =  params[1+1+1+1+1];
   int IA            =  params[1+1+1+1+1+1];
   int JA            =  params[1+1+1+1+1+1+1];
   int *DESCA        = &params[1+1+1+1+1+1+1+1];
   int IB            =  params[1+1+1+1+1+1+1+1+DLEN_];
   int JB            =  params[1+1+1+1+1+1+1+1+DLEN_+1];
   int *DESCB        = &params[1+1+1+1+1+1+1+1+DLEN_+1+1];
   hscaldoublereal BETA   =  params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_];
   int IC            =  params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1];
   int JC            =  params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1];
   int *DESCC        = &params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1];

   char TRANSAC   = TRANSAI;
   char TRANSA[2] = { TRANSAC, '\0' };

   char TRANSBC   = TRANSBI;
   char TRANSB[2] = { TRANSBC, '\0' };

   hscaldoublereal *a, *b, *c;

   int reconm            = M/p;
   int reconn            = N/q;
   int reconk            = hscal_pilaenv_recon("D");

   int lda, ldb, ldc;

   struct timeval start, end;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      gettimeofday(&start, NULL);
   }

   if (TRANSAC == 'N')
   {
      lda = reconm;
   }
   else
   {
      lda = reconk;
   }

   if (TRANSBC == 'N')
   {
      ldb = reconk;
   }
   else
   {
      ldb = reconn;
   }

   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "TRANSA %s TRANSB %s "
           "M %d N %d K %d "
           "reconm %d reconn %d reconk %d"
           "ALPHA %f "
           "IA %d JA %d IB %d JB %d "
           "BETA %f "
           "IC %d JC %d",
           TRANSA, TRANSB, 
           M, N, K, 
           reconm, reconn, reconk, 
           ALPHA,
           IA, JA, IB, JB, 
           BETA, 
           IC, JC
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return;
   }

   b = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return;
   }

   c = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i] = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i] = 1.;
   }

   //
   // Call the BLAS dgemm routine
   hscal_dgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      gettimeofday(&end, NULL);

      {
         double tstart  = start.tv_sec + start.tv_usec/1000000.;
         double tend    = end.tv_sec + end.tv_usec/1000000.;
         double t_sec   = tend - tstart;

         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            t_sec
         );
      }
   }

   return;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdgemm_modelp_t(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* TRANSA */
                  +        
                  1        /* TRANSB */
                  +
                  1        /* M */
                  +
                  1        /* N */
                  +        
                  1        /* K */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  1        /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *TRANSA;
   (*model_params)[1]                                               = *TRANSB;
   (*model_params)[1+1]                                             = *M;
   (*model_params)[1+1+1]                                           = *N;
   (*model_params)[1+1+1+1]                                         = *K;
   (*model_params)[1+1+1+1+1]                                       = alphai;
   (*model_params)[1+1+1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_]                           = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1]                         = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = betai;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]               = *IC;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]             = *JC;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCB[i];
       (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+i]     = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgemm_recon_t(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pdgemm_modelp_t(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon(
               &hscal_pdgemm_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*---------------------------------------------------------------------------*/

int main(int argc, char **argv)
{
    int        DESCA[DLEN_], DESCB[DLEN_], DESCC[DLEN_];
    int        i, rc;
    double     time;

    char TRANSAC = TRANSAI;
    char *TRANSA = &TRANSAC;
    char TRANSBC = TRANSBI;
    char *TRANSB = &TRANSBC;

    struct timeval gstart, gend;

    int num_of_processors;

    double algotime;

    MPI_Comm       *pdgemmcomm;

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (TRANSAC == 'N')
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[M_]     = M;
       DESCA[N_]     = K;
       DESCA[IMB_]   = MB;
       DESCA[INB_]   = KB;
       DESCA[MB_]    = MB;
       DESCA[NB_]    = KB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }
    else
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[M_]     = K;
       DESCA[N_]     = M;
       DESCA[IMB_]   = KB;
       DESCA[INB_]   = MB;
       DESCA[MB_]    = KB;
       DESCA[NB_]    = MB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }

    if (TRANSBC == 'N')
    {
       DESCB[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCB[M_]     = K;
       DESCB[N_]     = N;
       DESCB[IMB_]   = KB;
       DESCB[INB_]   = NB;
       DESCB[MB_]    = KB;
       DESCB[NB_]    = NB;
       DESCB[RSRC_]  = 0;
       DESCB[CSRC_]  = 0;
    }
    else
    {
       DESCB[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCB[M_]     = N;
       DESCB[N_]     = K;
       DESCB[IMB_]   = NB;
       DESCB[INB_]   = KB;
       DESCB[MB_]    = NB;
       DESCB[NB_]    = KB;
       DESCB[RSRC_]  = 0;
       DESCB[CSRC_]  = 0;
    }

    DESCC[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCC[M_]     = M;
    DESCC[N_]     = N;
    DESCC[IMB_]   = MB;
    DESCC[INB_]   = NB;
    DESCC[MB_]    = MB;
    DESCC[NB_]    = NB;
    DESCC[RSRC_]  = 0;
    DESCC[CSRC_]  = 0;

    num_of_processors = HMPI_Group_size(HMPI_COMM_WORLD_GROUP);
    p = q = sqrt(num_of_processors);

    for (i = 0; i < HSCAL_TEST_NUM_RECON_ITERATIONS; i++)
    {
        HMPI_Group gid;

        rc = hscal_pdgemm_recon_t(
                  TRANSA,
                  TRANSB,
                  &M,
                  &N,
                  &K,
                  &ALPHA,
                  &IA,
                  &JA,
                  DESCA,
                  &IB,
                  &JB,
                  DESCB,
                  &BETA,
                  &IC,
                  &JC,
                  DESCC
        );

        if (rc != HMPI_SUCCESS)
        {    
           return rc;
        }

        {
           int *model_params, model_count;

           if (HMPI_Is_host())
           {
              gettimeofday(&gstart, NULL);
           }

           rc = hscal_pack_pdgemm_modelp(
                 TRANSA,
                 TRANSB,
                 &M,
                 &N,
                 &K,
                 &ALPHA,
                 &IA,
                 &JA,
                 DESCA,
                 &IB,
                 &JB,
                 DESCB,
                 &BETA,
                 &IC,
                 &JC,
                 DESCC,
                 &model_params,
                 &model_count
           );

           if (rc != HSCAL_SUCCESS)
           {
              return rc;
           }

           HMPI_Debug(1);

           rc = HMPI_Group_heuristic_pauto_create(
                    &gid,
                    hscal_model_pdgemm,
                    hscal_heuristic_fn_all,
                    model_params,
                    model_count
           );

           if (rc != HMPI_SUCCESS)
           {
              free(model_params);
              return rc;
           }

           HMPI_Debug(0);

           free(model_params);

           if (HMPI_Is_host())
           {
              gettimeofday(&gend, NULL);
           }

/* 
* Broadcast the estimated execution time 
*/
           if (HMPI_Is_member(&gid))
           {
              int nc, *c;

              int plen;
              char pname[MPI_MAX_PROCESSOR_NAME];

              MPI_Get_processor_name(
                 pname,
                 &plen
              );
              
              algotime = HMPI_Group_timeof(&gid);

              HMPI_Group_coordof(
                  &gid,  
                  &nc,
                  &c
              );

              printf("RECON ITERATION %d: Coordinates (%d, %d), my node=%s\n", i+1, c[0], c[1], pname);

              free(c);
           }

/*
* Broadcast the optimal grid arrangement
*/
           if (HMPI_Is_host())
           {
              int nd;
              int *d;

              HMPI_Group_topology(
                  &gid,
                  &nd,
                  &d
              );

              p = d[0];

              free(d);

              rc = MPI_Bcast(
                      &p,
                      1,
                      MPI_INT,
                      0,
                      HMPI_COMM_WORLD
              );
           }
           else
           {
              rc = MPI_Bcast(
                      &p,
                      1,
                      MPI_INT,
                      0,
                      HMPI_COMM_WORLD
              );
           }

           if (rc != MPI_SUCCESS)
           {
              return HSCAL_ERR_BCAST_OPTIMAL_GRID;
           }

           if (HMPI_Is_host())
           {
              int nd;
              int *d;

              HMPI_Group_topology(
                  &gid,
                  &nd,
                  &d
              );

              q = d[1];

              free(d);

              rc = MPI_Bcast(
                      &q,
                      1,
                      MPI_INT,
                      0,
                      HMPI_COMM_WORLD
              );
           }
           else
           {
              rc = MPI_Bcast(
                      &q,
                      1,
                      MPI_INT,
                      0,
                      HMPI_COMM_WORLD
              );
           }

           if (rc != MPI_SUCCESS)
           {
              return HSCAL_ERR_BCAST_OPTIMAL_GRID;
           }

           if (HMPI_Is_parent(&gid))
           {
              printf("RECON Iteration finished %d: p=%d, q=%d, algotime=%f\n", i+1, p, q, algotime);
           }

           if (HMPI_Is_member(&gid))
           {
              HMPI_Group_free(
                  &gid
              );
           }
        }
    }

    HMPI_Finalize(0);
}

==> heteroscalapack-1.0.7-BETA.tar.gz: heteroscalapack-1.0.7-BETA/tests/RECON/level3/test_hscal_pdgemm_pauto.c <==

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <hscalapack.h>

#include "hscalapack_pdgemm_parameters.h"

#define  HSCAL_TEST_NUM_RECON_ITERATIONS  2

static int p, q;

/*------------------------------------------------------------------*/

void hscal_pdgemm_benchmark_function_t
(
     const void* input_p,
     int num_of_p,
     void* output_p
)
{
   int i, *params = (int*)input_p;

   int TRANSAI       =  params[0];
   int TRANSBI       =  params[1];
   int M             =  params[1+1];
   int N             =  params[1+1+1];
   int K             =  params[1+1+1+1];
   hscaldoublereal ALPHA  =  params[1+1+1+1+1];
   int IA            =  params[1+1+1+1+1+1];
   int JA            =  params[1+1+1+1+1+1+1];
   int *DESCA        = &params[1+1+1+1+1+1+1+1];
   int IB            =  params[1+1+1+1+1+1+1+1+DLEN_];
   int JB            =  params[1+1+1+1+1+1+1+1+DLEN_+1];
   int *DESCB        = &params[1+1+1+1+1+1+1+1+DLEN_+1+1];
   hscaldoublereal BETA   =  params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_];
   int IC            =  params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1];
   int JC            =  params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1];
   int *DESCC        = &params[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1];

   char TRANSAC   = TRANSAI;
   char TRANSA[2] = { TRANSAC, '\0' };

   char TRANSBC   = TRANSBI;
   char TRANSB[2] = { TRANSBC, '\0' };

   hscaldoublereal *a, *b, *c;

   int reconm            = M/p;
   int reconn            = N/q;
   int reconk            = hscal_pilaenv_recon("D");

   int lda, ldb, ldc;

   struct timeval start, end;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      gettimeofday(&start, NULL);
   }

   if (TRANSAC == 'N')
   {
      lda = reconm;
   }
   else
   {
      lda = reconk;
   }

   if (TRANSBC == 'N')
   {
      ldb = reconk;
   }
   else
   {
      ldb = reconn;
   }

   ldc = reconm;

   if (HMPI_Is_host())
   {
      HSCAL_DEBUG(
           __FILE__, __FUNCTION__,__LINE__,
           "TRANSA %s TRANSB %s "
           "M %d N %d K %d "
           "reconm %d reconn %d reconk %d"
           "ALPHA %f "
           "IA %d JA %d IB %d JB %d "
           "BETA %f "
           "IC %d JC %d",
           TRANSA, TRANSB, 
           M, N, K, 
           reconm, reconn, reconk, 
           ALPHA,
           IA, JA, IB, JB, 
           BETA, 
           IC, JC
      );
   }

   //
   // Allocate the matrices A, B, and C
   a = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconm*reconk)
   );

   if (a == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return;
   }

   b = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconk*reconn)
   );

   if (b == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return;
   }

   c = (hscaldoublereal*)malloc(
                 sizeof(hscaldoublereal)
                 *
                 (reconm*reconn)
   );

   if (c == NULL)
   {
      *(int*)output_p = HSCAL_ERR_LOW_MEMORY;
      return;
   }

   //
   // Initialize the matrices A, B, and C
   for (i = 0; i < (reconm*reconk); i++)
   {
      a[i] = 2.;
   }

   for (i = 0; i < (reconk*reconn); i++)
   {
      b[i] = 2.;
   }

   for (i = 0; i < (reconm*reconn); i++)
   {
      c[i] = 1.;
   }

   //
   // Call the BLAS dgemm routine
   hscal_dgemm_benchmark(
        TRANSA,
        TRANSB,
        &reconm,
        &reconn,
        &reconk,
        &ALPHA,
        a,
        &lda,
        b,
        &ldb,
        &BETA,
        c,
        &ldc
   );

   //
   // Free the matrices A, B, and C
   free(a);
   free(b);
   free(c);

   *(int*)output_p = HSCAL_SUCCESS;

   if (hscal_debug_flag & HSCAL_LOG_TIMINGS)
   {
      gettimeofday(&end, NULL);

      {
         double tstart  = start.tv_sec + start.tv_usec/1000000.;
         double tend    = end.tv_sec + end.tv_usec/1000000.;
         double t_sec   = tend - tstart;

         HSCAL_TIMINGS(
            "me=%d, reconm=%d, reconn=%d, reconk=%d, time(sec)=%f",
            HMPI_Group_rank(HMPI_COMM_WORLD_GROUP),
            reconm,
            reconn,
            reconk,
            t_sec
         );
      }
   }

   return;
}

/*------------------------------------------------------------------*/

int hscal_pack_pdgemm_modelp_t(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC,
   int **model_params,
   int *model_count
)
{
   int i;

   /*
   * Just get a indicative multiplication scalar of 
   * alpha and beta
   */
   int alphai, betai;

   if (*ALPHA == 0.)
   {
      alphai = 0;
   }
   else if (*ALPHA == 1.)
   {
      alphai = 1;
   }
   else
   {
      alphai = 2;
   }

   if (*BETA == 0.)
   {
      betai = 0;
   }
   else if (*BETA == 1.)
   {
      betai = 1;
   }
   else
   {
      betai = 2;
   }

   *model_count = 1        /* TRANSA */
                  +        
                  1        /* TRANSB */
                  +
                  1        /* M */
                  +
                  1        /* N */
                  +        
                  1        /* K */
                  +
                  1        /* ALPHA */
                  +
                  1        /* IA */
                  +
                  1        /* JA */ 
                  +
                  DLEN_    /* DESCA */
                  +
                  1        /* IB */
                  +
                  1        /* JB */ 
                  +
                  DLEN_    /* DESCB */
                  +
                  1        /* BETA */
                  +
                  1        /* IC */
                  +
                  1        /* JC */
                  +
                  DLEN_    /* DESCC */
   ;

   *model_params = (int*)malloc(
                         (*model_count)*
                         sizeof(int)
   );

   if (*model_params == NULL)
   {
      return MPC_ERR_NOMEM;
   }

   (*model_params)[0]                                               = *TRANSA;
   (*model_params)[1]                                               = *TRANSB;
   (*model_params)[1+1]                                             = *M;
   (*model_params)[1+1+1]                                           = *N;
   (*model_params)[1+1+1+1]                                         = *K;
   (*model_params)[1+1+1+1+1]                                       = alphai;
   (*model_params)[1+1+1+1+1+1]                                     = *IA;
   (*model_params)[1+1+1+1+1+1+1]                                   = *JA;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_]                           = *IB;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1]                         = *JB;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_]                 = betai;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1]               = *IC;
   (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1]             = *JC;

   for (i = 0; i < DLEN_; i++)
   {
       (*model_params)[1+1+1+1+1+1+1+1+i]                           = DESCA[i];
       (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+i]                 = DESCB[i];
       (*model_params)[1+1+1+1+1+1+1+1+DLEN_+1+1+DLEN_+1+1+1+i]     = DESCC[i];
   }

   return HSCAL_SUCCESS;
}

/*------------------------------------------------------------------*/

int hscal_pdgemm_recon_t(
   F_CHAR_T TRANSA, 
   F_CHAR_T TRANSB,
   int * M,
   int * N, 
   int * K,
   double * ALPHA,
   int * IA, 
   int * JA, 
   int * DESCA,
   int * IB, 
   int * JB, 
   int * DESCB,
   double * BETA,
   int * IC, 
   int * JC, 
   int * DESCC
)
{
   int rc;

   if (HMPI_Is_member(HMPI_COMM_WORLD_GROUP))
   {
      int output_p;
      int *model_params, model_count;

      rc = hscal_pack_pdgemm_modelp_t(
                TRANSA,
                TRANSB,
                M,
                N,
                K,
                ALPHA,
                IA,
                JA,
                DESCA,
                IB,
                JB,
                DESCB,
                BETA,
                IC,
                JC,
                DESCC,
                &model_params,
                &model_count  
      );      

      if (rc != HSCAL_SUCCESS)
      {
         return rc;
      }

      rc = HMPI_Recon(
               &hscal_pdgemm_benchmark_function,
               model_params,
               model_count,
               &output_p
      );

      if (rc != HMPI_SUCCESS)
      {
         free(model_params);
         return rc;
      }

      if (HMPI_Is_member(HMPI_PROC_WORLD_GROUP))
      {
         if (output_p != HSCAL_SUCCESS)
         {
            free(model_params);
            return output_p;
         }
      }

      free(model_params);        
   }

   return HSCAL_SUCCESS;
}

/*---------------------------------------------------------------------------*/

int main(int argc, char **argv)
{
    int        DESCA[DLEN_], DESCB[DLEN_], DESCC[DLEN_];
    int        i, rc;
    double     time;

    char TRANSAC = TRANSAI;
    char *TRANSA = &TRANSAC;
    char TRANSBC = TRANSBI;
    char *TRANSB = &TRANSBC;

    struct timeval gstart, gend;

    int num_of_processors;

    double algotime;

    MPI_Comm       pdgemmcomm;

    //
    // Initialize HMPI runtime
    HMPI_Init(&argc, &argv);

    if (TRANSAC == 'N')
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[M_]     = M;
       DESCA[N_]     = K;
       DESCA[IMB_]   = MB;
       DESCA[INB_]   = KB;
       DESCA[MB_]    = MB;
       DESCA[NB_]    = KB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }
    else
    {
       DESCA[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCA[M_]     = K;
       DESCA[N_]     = M;
       DESCA[IMB_]   = KB;
       DESCA[INB_]   = MB;
       DESCA[MB_]    = KB;
       DESCA[NB_]    = MB;
       DESCA[RSRC_]  = 0;
       DESCA[CSRC_]  = 0;
    }

    if (TRANSBC == 'N')
    {
       DESCB[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCB[M_]     = K;
       DESCB[N_]     = N;
       DESCB[IMB_]   = KB;
       DESCB[INB_]   = NB;
       DESCB[MB_]    = KB;
       DESCB[NB_]    = NB;
       DESCB[RSRC_]  = 0;
       DESCB[CSRC_]  = 0;
    }
    else
    {
       DESCB[DTYPE_] = BLOCK_CYCLIC_2D_INB;
       DESCB[M_]     = N;
       DESCB[N_]     = K;
       DESCB[IMB_]   = NB;
       DESCB[INB_]   = KB;
       DESCB[MB_]    = NB;
       DESCB[NB_]    = KB;
       DESCB[RSRC_]  = 0;
       DESCB[CSRC_]  = 0;
    }

    DESCC[DTYPE_] = BLOCK_CYCLIC_2D_INB;
    DESCC[M_]     = M;
    DESCC[N_]     = N;
    DESCC[IMB_]   = MB;
    DESCC[INB_]   = NB;
    DESCC[MB_]    = MB;
    DESCC[NB_]    = NB;
    DESCC[RSRC_]  = 0;
    DESCC[CSRC_]  = 0;

    num_of_processors = HMPI_Group_size(HMPI_COMM_WORLD_GROUP);
    p = q = sqrt(num_of_processors);

    for (i = 0; i < HSCAL_TEST_NUM_RECON_ITERATIONS; i++)
    {
        HMPI_Group gid;

        rc = hscal_pdgemm_recon_t(
                  TRANSA,
                  TRANSB,
                  &M,
                  &N,
                  &K,
                  &ALPHA,
                  &IA,
                  &JA,
                  DESCA,
                  &IB,
                  &JB,
                  DESCB,
                  &BETA,
                  &IC,
                  &JC,
                  DESCC
        );

        if (rc != HMPI_SUCCESS)
        {    
           return rc;
        }

        {
           int *model_params, model_count;

           if (HMPI_Is_host())
           {
              gettimeofday(&gstart, NULL);
           }

           rc = hscal_pack_pdgemm_modelp(
                 TRANSA,
                 TRANSB,
                 &M,
                 &N,
                 &K,
                 &ALPHA,
                 &IA,
                 &JA,
                 DESCA,
                 &IB,
                 &JB,
                 DESCB,
                 &BETA,
                 &IC,
                 &JC,
                 DESCC,
                 &model_params,
                 &model_count
           );

           if (rc != HSCAL_SUCCESS)
           {
              return rc;
           }

           HMPI_Debug(1);

           rc = HMPI_Group_heuristic_pauto_create(
                    &gid,
                    hscal_model_pdgemm,
                    hscal_heuristic_fn_square,
                    model_params,
                    model_count
           );

           if (rc != HMPI_SUCCESS)
           {
              free(model_params);
              return rc;
           }

           HMPI_Debug(0);

           free(model_params);

           if (HMPI_Is_host())
           {
              gettimeofday(&gend, NULL);
           }

           if (HMPI_Is_free())
           {
              continue;
           }

           pdgemmcomm = *(MPI_Comm*)HMPI_Get_comm(&gid);

/* 
* Broadcast the estimated execution time 
*/
           if (HMPI_Is_member(&gid))
           {
              int nc, *c;

              int plen;
              char pname[MPI_MAX_PROCESSOR_NAME];

              MPI_Get_processor_name(
                 pname,
                 &plen
              );
              
              algotime = HMPI_Group_timeof(&gid);

              HMPI_Group_coordof(
                  &gid,  
                  &nc,
                  &c
              );

              printf("RECON ITERATION %d: Coordinates (%d, %d), my node=%s\n", i+1, c[0], c[1], pname);

              free(c);
           }

/*
* Broadcast the optimal grid arrangement
*/
           if (HMPI_Is_parent(&gid))
           {
              int nd;
              int *d;

              HMPI_Group_topology(
                  &gid,
                  &nd,
                  &d
              );

              p = d[0];

              free(d);

              rc = MPI_Bcast(
                      &p,
                      1,
                      MPI_INT,
                      0,
                      pdgemmcomm
              );
           }
           else
           {
              rc = MPI_Bcast(
                      &p,
                      1,
                      MPI_INT,
                      0,
                      pdgemmcomm
              );
           }

           if (rc != MPI_SUCCESS)
           {
              return HSCAL_ERR_BCAST_OPTIMAL_GRID;
           }

           if (HMPI_Is_parent(&gid))
           {
              int nd;
              int *d;

              HMPI_Group_topology(
                  &gid,
                  &nd,
                  &d
              );

              q = d[1];

              free(d);

              rc = MPI_Bcast(
                      &q,
                      1,
                      MPI_INT,
                      0,
                      pdgemmcomm
              );
           }
           else
           {
              rc = MPI_Bcast(
                      &q,
                      1,
                      MPI_INT,
                      0,
                      pdgemmcomm
              );
           }

           if (rc != MPI_SUCCESS)
           {
              return HSCAL_ERR_BCAST_OPTIMAL_GRID;
           }

           if (HMPI_Is_parent(&gid))
           {
              printf("RECON Iteration finished %d: p=%d, q=%d, algotime=%f\n", i+1, p, q, algotime);
           }

           if (HMPI_Is_member(&gid))
           {
              HMPI_Group_free(
                  &gid
              );
           }
        }
    }

    HMPI_Finalize(0);
}
